'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _cliEngineCommand = require('cli-engine-command');

var _cliEngineCommand2 = _interopRequireDefault(_cliEngineCommand);

var _plugins = require('./plugins');

var _plugins2 = _interopRequireDefault(_plugins);

var _updater = require('./updater');

var _updater2 = _interopRequireDefault(_updater);

var _not_found = require('./not_found');

var _not_found2 = _interopRequireDefault(_not_found);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

if (process.env.HEROKU_TIME_REQUIRE) require('time-require');

const handleEPIPE = err => {
  if (err.code !== 'EPIPE') throw err;
};

class Main extends _cliEngineCommand2.default {
  constructor(options) {
    super(new _cliEngineCommand.Config(options));
  }

  async run() {
    process.on('exit', () => this.showCursor());
    process.on('SIGINT', () => this.exit(1));
    process.on('uncaughtException', err => this.error(err));
    process.on('unhandledRejection', err => this.error(err));
    process.stdout.on('error', handleEPIPE);
    process.stderr.on('error', handleEPIPE);

    const updater = new _updater2.default(this.config);
    const plugins = new _plugins2.default(this.config);
    await updater.autoupdate();
    await plugins.refreshLinkedPlugins();
    let Command = plugins.findCommand(this.config.argv[1] || this.config.defaultCommand);
    if (!Command) return new _not_found2.default(this.config).run();
    await Command.run(this.config.argv.slice(2), { config: this.config });
    this.exit(0);
  }
}
exports.default = Main;