'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _cliEngineCommand = require('cli-engine-command');

var _cliEngineCommand2 = _interopRequireDefault(_cliEngineCommand);

var _util = require('../util');

var _util2 = _interopRequireDefault(_util);

var _screen = require('cli-engine-command/lib/output/screen');

var _plugins = require('../plugins');

var _plugins2 = _interopRequireDefault(_plugins);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Help extends _cliEngineCommand2.default {

  async run() {
    this.plugins = new _plugins2.default(this.config);
    let cmd = this.argv.find(arg => !['help', '-h', '--help'].includes(arg));
    if (!cmd) return this.topics();
    let Topic = this.plugins.findTopic(cmd);
    let matchedCommand = this.plugins.findCommand(cmd);
    if (!Topic) throw new Error(`command ${cmd} not found`);
    let commands = this.plugins.commandsForTopic(Topic.topic);
    await new Topic(commands, this).help(this.argv, matchedCommand);
  }

  topics() {
    this.log(`Usage: ${this.config.bin} COMMAND [--app APP] [command-specific-options]

Help topics, type ${this.color.cmd(this.config.bin + ' help TOPIC')} for more details:\n`);
    let topics = this.plugins.topics.filter(t => !t.hidden);
    topics.sort(_util2.default.compare('topic'));
    topics = topics.map(t => [t.topic, t.description]);
    this.log(this.renderList(topics));
    this.log();
  }

  renderList(items) {
    const S = require('string');
    const max = require('lodash.maxby');

    let maxLength = max(items, '[0].length')[0].length + 1;
    let lines = items.map(i => [
    // left side
    ` ${S(i[0]).padRight(maxLength)}`,

    // right side
    this.linewrap(maxLength + 4, i[1] || '')])
    // join left + right side
    .map(i => i[1] ? `${i[0]} # ${i[1]}` : i[0]);
    return lines.join('\n');
  }

  linewrap(length, s) {
    const linewrap = require('../linewrap');
    return linewrap(length, _screen.stdtermwidth, {
      skipScheme: 'ansi-color'
    })(s).trim();
  }
}
exports.default = Help;
Help.topic = 'help';
Help.description = 'display help';
Help.variableArgs = true;