'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Plugin = undefined;

var _cliEngineCommand = require('cli-engine-command');

var _cliEngineCommand2 = _interopRequireDefault(_cliEngineCommand);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _yarn = require('./yarn');

var _yarn2 = _interopRequireDefault(_yarn);

var _rwlockfile = require('rwlockfile');

var _rwlockfile2 = _interopRequireDefault(_rwlockfile);

var _linked_plugins = require('./linked_plugins');

var _linked_plugins2 = _interopRequireDefault(_linked_plugins);

var _lodash = require('lodash.uniqby');

var _lodash2 = _interopRequireDefault(_lodash);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Cache extends _cliEngineCommand.Base {

  get file() {
    return _path2.default.join(this.config.dirs.cache, 'plugins.json');
  }
  get cache() {
    if (this._cache) return this._cache;
    let initial = { version: this.config.version, plugins: {} };
    try {
      this._cache = this.fs.readJSONSync(this.file);
    } catch (err) {
      if (err.code !== 'ENOENT') throw err;
      this._cache = initial;
    }
    if (this._cache.version !== this.config.version) this._cache = initial;
    return this._cache;
  }

  plugin(path) {
    return this.cache.plugins[path];
  }

  updatePlugin(path, plugin) {
    this.constructor.updated = true;
    this.cache.plugins[path] = plugin;
  }

  deletePlugin(name) {
    for (let k of Object.keys(this.cache.plugins)) {
      if (this.cache.plugins[k].name === name) {
        this.debug(`Clearing cache for ${k}`);
        this.constructor.updated = true;
        delete this.cache.plugins[k];
      }
    }
    this.save();
  }

  save() {
    if (!this.constructor.updated) return;
    try {
      this.fs.writeJSONSync(this.file, this.cache);
    } catch (err) {
      this.warn(err);
    }
  }
}

Cache.updated = false;
function undefaultTopic(t) {
  if (t.default) return t.default;
  return t;
}

function undefaultCommand(c) {
  if (c.default && typeof c.default !== 'boolean') return c.default;
  return c;
}

class Plugin extends _cliEngineCommand.Base {
  constructor(type, path, config, cache) {
    super(config);
    this.cache = cache;
    this.type = type;
    this.path = path;
    let p = this.fetch();
    this.name = p.name;
    this.version = p.version;
  }

  get commands() {
    return this.fetch().commands;
  }

  get topics() {
    return this.fetch().topics;
  }

  findCommand(cmd) {
    let c = this.commands.find(c => c.id === cmd || (c.aliases || []).includes(cmd));
    if (!c) return;
    let { topic, command } = c;
    let p = this.require();
    let Command = (p.commands || []).map(undefaultCommand).find(d => topic === d.topic && command === d.command);
    if (!Command) return;
    return typeof Command === 'function' ? Command : this.buildCommand(Command);
  }

  findTopic(name) {
    let t = this.topics.find(t => t.topic === name);
    if (!t) return;
    let Topic = (this.require().topics || []).find(t => [t.topic, t.name].includes(name));
    if (!Topic) return;
    return typeof Topic === 'function' ? Topic : this.buildTopic(t);
  }

  buildTopic(t) {
    var _class, _temp;

    return _temp = _class = class extends _cliEngineCommand.Topic {}, _class.topic = t.topic, _class.description = t.description, _class.hidden = t.hidden, _temp;
  }

  buildCommand(c) {
    var _class2, _temp2;

    if (!c.topic) throw new Error('command has no topic');
    let Base = c.needsApp || c.wantsApp ? _cliEngineCommand.mixins.app(_cliEngineCommand2.default, { required: !!c.needsApp }) : _cliEngineCommand2.default;
    return _temp2 = _class2 = class extends Base {

      run() {
        const ctx = {
          supportsColor: this.color.enabled,
          auth: {},
          debug: this.config.debug,
          flags: this.flags,
          args: c.variableArgs ? this.argv : this.args,
          // flow$ignore
          app: this.app
        };
        if (c.needsAuth) {
          ctx.auth.password = process.env.HEROKU_API_KEY;
          if (!ctx.auth.password) {
            const netrc = require('netrc')();
            const host = netrc['api.heroku.com'];
            if (host) ctx.auth.password = host.password;
          }
          if (!ctx.auth.password) throw new Error('Not logged in');
        }
        return c.run(ctx);
      }
    }, _class2.topic = c.topic, _class2.command = c.command, _class2.description = c.description, _class2.hidden = c.hidden, _class2.args = c.args || [], _class2.flags = c.flags || [], _class2.variableArgs = c.variableArgs, _temp2;
  }

  fetch() {
    let c = this.cache.plugin(this.path);
    if (c) return c;
    try {
      this.debug('updating cache for ' + this.path);
      return this.updatePlugin(this.require());
    } catch (err) {
      if (this.type === 'builtin') throw err;
      this.warn(err);
      return {
        name: this.path,
        path: this.path,
        version: '',
        commands: [],
        topics: []
      };
    }
  }

  updatePlugin(plugin) {
    const name = this.type === 'builtin' ? 'builtin' : this.pjson().name;
    const version = this.type === 'builtin' ? this.config.version : this.pjson().version;
    if (!plugin.commands) throw new Error('no commands found');
    const commands = plugin.commands.map(undefaultCommand).map(c => ({
      id: c.command ? `${c.topic}:${c.command}` : c.topic,
      topic: c.topic,
      command: c.command,
      description: c.description,
      args: c.args,
      variableArgs: c.variableArgs,
      flags: c.flags,
      help: c.help,
      usage: c.usage,
      hidden: !!c.hidden,
      aliases: c.aliases
    }));
    const topics = (plugin.topics || (plugin.topic ? [plugin.topic] : [])).map(undefaultTopic).map(t => ({
      topic: t.topic || t.name || '',
      description: t.description,
      hidden: !!t.hidden
    }));

    for (let command of commands) {
      if (topics.find(t => t.topic === command.topic)) continue;
      topics.push({
        topic: command.topic,
        hidden: true
      });
    }

    const cachedPlugin = { name, path: this.path, version, commands, topics };
    this.cache.updatePlugin(this.path, cachedPlugin);
    return cachedPlugin;
  }

  // flow$ignore
  require() {
    return require(this.path);
  }
  // flow$ignore
  pjson() {
    return require(_path2.default.join(this.path, 'package.json'));
  }
}

exports.Plugin = Plugin;
class Plugins extends _cliEngineCommand.Base {
  constructor(config) {
    super(config);
    this.config = config;
    this.cache = new Cache(config);
    this.linkedPlugins = new _linked_plugins2.default(config, this);
    this.plugins = [new Plugin('builtin', './commands', config, this.cache)].concat(this.linkedPlugins.list()).concat(this.userPlugins).concat(this.corePlugins);
    this.cache.save();
    this.yarn = new _yarn2.default(this.config);
  }

  get corePlugins() {
    return (this.config._cli.plugins || []).map(name => {
      return new Plugin('core', _path2.default.join(this.config.root, 'node_modules', name), this.config, this.cache);
    });
  }

  get userPlugins() {
    const pjson = this.userPluginsPJSON;
    return Object.keys(pjson.dependencies || {}).map(name => {
      return new Plugin('user', this.userPluginPath(name), this.config, this.cache);
    });
  }

  get userPluginsPJSON() {
    try {
      return this.fs.readJSONSync(_path2.default.join(this.userPluginsDir, 'package.json'));
    } catch (err) {
      return { dependencies: {} };
    }
  }

  list() {
    return this.plugins;
  }

  isPluginInstalled(name) {
    return !!this.plugins.find(p => p.name === name);
  }

  findCommand(cmd) {
    for (let plugin of this.plugins) {
      let c = plugin.findCommand(cmd);
      if (c) return c;
    }
  }

  commandsForTopic(topic) {
    let commands = this.plugins.reduce((t, p) => {
      return t.concat(p.commands.filter(c => c.topic === topic).map(c => p.findCommand(c.id)));
    }, []);
    return (0, _lodash2.default)(commands, 'id');
  }

  findTopic(cmd) {
    let name = cmd.split(':')[0];
    for (let plugin of this.plugins) {
      let t = plugin.findTopic(name);
      if (t) return t;
    }
  }

  async setupUserPlugins() {
    const pjson = _path2.default.join(this.userPluginsDir, 'package.json');
    const yarnrc = _path2.default.join(this.userPluginsDir, '.yarnrc');
    this.fs.mkdirpSync(this.userPluginsDir);
    if (!this.fs.existsSync(pjson)) this.fs.writeFileSync(pjson, JSON.stringify({ private: true }));
    if (!this.fs.existsSync(yarnrc)) this.fs.writeFileSync(yarnrc, 'registry "https://cli-npm.heroku.com/"');
    await this.yarn.exec();
  }

  async install(name) {
    let unlock = await _rwlockfile2.default.write(this.lockfile, { skipOwnPid: true });
    await this.setupUserPlugins();
    if (this.plugins.find(p => p.name === name)) throw new Error(`Plugin ${name} is already installed`);
    if (!this.config.debug) this.action.start(`Installing plugin ${name}`);
    await this.yarn.exec(['add', name]);
    this.clearCache(name);
    try {
      // flow$ignore
      let plugin = require(this.userPluginPath(name));
      if (!plugin.commands) throw new Error(`${name} does not appear to be a Heroku CLI plugin`);
    } catch (err) {
      await unlock();
      this.error(err, false);
      await this.uninstall(name);
      this.exit(1);
    }
    this.action.stop();
    await unlock();
  }

  /**
   * check if a plugin needs an update
   * @param {string} name - plugin name to check
   * @returns {string} - latest version if needs update, undefined otherwise
   */
  async needsUpdate(name) {
    let info = await this.yarn.info(name);
    let plugin = this.plugins.find(p => p.name === name);
    if (!plugin) throw new Error(`${name} not installed`);
    let latest = info['dist-tags'].latest;
    if (latest === plugin.version) return;
    return latest;
  }

  async update(name, version) {
    await this.yarn.exec(['upgrade', `${name}@${version}`]);
    this.clearCache(name);
  }

  async uninstall(name) {
    let unlock = await _rwlockfile2.default.write(this.lockfile, { skipOwnPid: true });
    let plugin = this.plugins.filter(p => !['core', 'builtin'].includes(p.type)).find(p => p.name === name);
    if (!plugin) throw new Error(`${name} is not installed`);
    switch (plugin.type) {
      case 'user':
        {
          if (!this.config.debug) this.action.start(`Uninstalling plugin ${name}`);
          await this.yarn.exec(['remove', name]);
          break;
        }
      case 'link':
        {
          if (!this.config.debug) this.action.start(`Unlinking plugin ${name}`);
          this.linkedPlugins.remove(plugin.path);
          break;
        }
    }
    this.clearCache(name);
    await unlock();
    this.action.stop();
  }

  async addLinkedPlugin(p) {
    await this.linkedPlugins.add(p);
  }

  async refreshLinkedPlugins() {
    await this.linkedPlugins.refresh();
  }

  clearCache(name) {
    this.cache.deletePlugin(name);
  }

  get userPluginsDir() {
    return _path2.default.join(this.config.dirs.data, 'plugins');
  }
  userPluginPath(name) {
    return _path2.default.join(this.userPluginsDir, 'node_modules', name);
  }

  get topics() {
    return (0, _lodash2.default)(this.plugins.reduce((t, p) => t.concat(p.topics), []), 'topic');
  }

  get lockfile() {
    return _path2.default.join(this.config.dirs.cache, 'plugins.lock');
  }

}
exports.default = Plugins;