'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _cliEngineCommand = require('cli-engine-command');

var _rwlockfile = require('rwlockfile');

var _rwlockfile2 = _interopRequireDefault(_rwlockfile);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Yarn extends _cliEngineCommand.Base {
  get lockfile() {
    return _path2.default.join(this.config.dirs.cache, 'yarn.lock');
  }

  get nodeModulesDirs() {
    return require('find-node-modules')({ cwd: __dirname, relative: false });
  }
  get yarnDir() {
    return this.nodeModulesDirs.map(d => _path2.default.join(d, 'yarn')).find(f => this.fs.existsSync(f));
  }
  get bin() {
    return this.yarnDir ? _path2.default.join(this.yarnDir, 'bin', 'yarn') : 'yarn';
  }

  async exec(args = [], options = {}) {
    let deleteYarnRoadrunnerCache = () => {
      let getDirectory = category => {
        // use %LOCALAPPDATA%/Yarn on Windows
        if (process.platform === 'win32' && process.env.LOCALAPPDATA) {
          return _path2.default.join(process.env.LOCALAPPDATA, 'Yarn', category);
        }

        // otherwise use ~/.{category}/yarn
        return _path2.default.join(this.config.dirs.home, `.${category}`, 'yarn');
      };

      let getCacheDirectory = () => {
        if (process.platform === 'darwin') {
          return _path2.default.join(this.config.dirs.home, 'Library', 'Caches', 'Yarn');
        }

        return getDirectory('cache');
      };

      try {
        this.fs.unlinkSync(_path2.default.join(getCacheDirectory(), '.roadrunner.json'));
      } catch (err) {}
    };

    options = Object.assign({
      cwd: _path2.default.join(this.config.dirs.data, 'plugins'),
      stdio: this.config.debug ? 'inherit' : null
    }, options);

    if (!this.bin) throw new Error('yarn not found');
    const execa = require('execa');
    this.debug(`${options.cwd}: ${this.bin} ${args.join(' ')}`);
    deleteYarnRoadrunnerCache();
    let unlock = await _rwlockfile2.default.write(this.lockfile);
    let output = await execa(this.bin, args, options);
    await unlock();
    deleteYarnRoadrunnerCache();
    return output.stdout;
  }

  async info(name) {
    let output = await this.exec(['info', name, '--json'], { stdio: null });
    let info = JSON.parse(output)['data'];
    return info;
  }
}
exports.default = Yarn;